# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Decorate ``check`` command, improving error handling."""

from django.core.management.commands import check

from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand, check.Command):
    """Decorate ``check`` command, improving error handling."""
