#!/bin/sh

set -e

# Ensure that recent versions of important dependencies will be installed from
# trixie-backports. This is a no-op if trixie-backports isn't in the APT
# sources.
# (This is a placeholder; there are no required backports right now.)
# cat >/etc/apt/preferences.d/99backports <<EOT
# Package: ...
# Pin: release n=bookworm-backports
# Pin-Priority: 900
# EOT
