{% load debusine %}
{% with workflow_root=work_request.workflow_root %}
    {% if workflow_root %}
        {% with collection=workflow_root.internal_collection %}
            <p id="internal-collection">
                Internal collection: <a href="{{ collection.get_absolute_url }}">{{ collection.name }}</a>
            </p>
        {% endwith %}
    {% endif %}
{% endwith %}
<h2>Task data</h2>
{% include "web/_card.html" with id="card-task_data" data=task_data only %}
<h2>Task data after applying task configuration</h2>
{% if show_task_data_configured %}
    {% include "web/_card.html" with id="card-configured_task_data" data=task_data_configured only %}
{% else %}
    <p id="task_data_not_configured">Task has not yet been configured.</p>
{% endif %}
<h2>Scheduler tags</h2>
{% if show_task_data_configured %}
    {% include "web/_card.html" with id="card-scheduler-tags" data=scheduler_tags only %}
{% else %}
    <p id="task_data_not_configured1">Task has not yet been configured.</p>
{% endif %}
<h2>Dynamic data</h2>
{% if show_task_data_dynamic %}
    {% include "web/_card.html" with id="card-dynamic_data" data=task_data_dynamic only %}
{% else %}
    <p id="dynamic_data_not_available">Dynamic task data has not yet been computed.</p>
{% endif %}
<h2>Workflow data</h2>
{% include "web/_card.html" with id="card-workflow_data" data=work_request_workflow_data_json only %}
<h2>Event reactions</h2>
{% include "web/_card.html" with id="card-event_reactions" data=work_request_event_reactions_json only %}
<h2>Output data</h2>
{% if work_request_output_data %}
    {% include "web/_card.html" with id="card-output_data" data=work_request_output_data_json only %}
{% else %}
    <p id="internals-no-output-data">No output data.</p>
{% endif %}
