{% extends base_template %}
{% load debusine %}
{% block content %}
    <div id="suite-summary">
        <p>{{ suite.name }}: {{ suite.description|default_if_none:"" }}</p>
    </div>
    {% if suite.architectures %}
        <div id="suite-architectures">
            <p>Architectures:</p>
            <ul>
                {% for architecture in suite.architectures %}<li>{{ architecture }}</li>{% endfor %}
            </ul>
        </div>
    {% endif %}
    {% if suite.release_fields %}
        <div id="suite-release-fields">
            <p>Release fields:</p>
            <dl>
                {% for name, value in suite.release_fields.items %}
                    <dt>{{ name }}</dt>
                    <dd>
                        {{ value }}
                    </dd>
                {% endfor %}
            </dl>
        </div>
    {% endif %}
    <div id="suite-apt">
        <p>APT source:</p>
        <pre><code>{{ suite.as_deb822_source }}</code></pre>
    </div>
{% endblock %}
